indexing

	description: "Base class that covers certain file descriptor routines."

	idea: "Functions here are also available under Windows more or less.%
	%This also means that an abstract file descriptor cannot be converted to a%
	%stream, because Windows doesn't support that."

	author: "Berend de Boer"
	date: "$Date: 2003/01/18 $"
	revision: "$Revision: #4 $"


deferred class

	ABSTRACT_FILE_DESCRIPTOR


inherit

	STDC_BASE

	EPX_POINTER_ADD

	MEMORY
		export
			{NONE} all
		redefine
			dispose
		end

	PORTABLE_PATH
		rename
			portable_path as path
		export
			{ANY} path
		end


feature -- Creation

	open (a_path: STRING; a_flags: INTEGER) is
			-- open given file with access given by `flags'
		require
			closed: is_closed
		do
			make_path
			do_open (a_path, a_flags)
		ensure
			is_opened: security.error_handling.exceptions_enabled implies is_open
		end

	open_read (a_path: STRING) is
			-- open given file with read-only access
		require
			closed: is_closed
		do
			open (a_path, abstract_O_RDONLY)
		ensure
			is_opened: security.error_handling.exceptions_enabled implies is_open
		end

	open_write (a_path: STRING) is
		require
			closed: is_closed
		do
			open (a_path, abstract_O_WRONLY)
		ensure
			is_opened: security.error_handling.exceptions_enabled implies is_open
		end

	open_read_write (a_path: STRING) is
		require
			closed: is_closed
		do
			open (a_path, abstract_O_RDWR)
		ensure
			is_opened: security.error_handling.exceptions_enabled implies is_open
		end

	open_truncate (a_path: STRING) is
		require
			closed: is_closed
		do
			open (a_path, abstract_O_TRUNC)
		ensure
			is_opened: security.error_handling.exceptions_enabled implies is_open
		end

	create_read_write (a_path: STRING) is
			-- Always create a file, existing or not.
			-- Give read/write permissions to user only.
		require
			closed: is_closed
		do
			create_with_mode (
				a_path,
				abstract_O_CREAT + abstract_O_TRUNC + abstract_O_RDWR,
				abstract_S_IREAD + abstract_S_IWRITE)
		ensure
			is_opened: security.error_handling.exceptions_enabled implies is_open
			-- only user has access (read/write) permissions
		end

	create_write (a_path: STRING) is
			-- Always create a file, existing or not.
			-- Give read/write permissions to user only.
		require
			closed: is_closed
		do
			create_with_mode (
				a_path,
				abstract_O_CREAT + abstract_O_TRUNC + abstract_O_WRONLY,
				abstract_S_IREAD + abstract_S_IWRITE)
		ensure
			is_opened: security.error_handling.exceptions_enabled implies is_open
			-- only user has access (read/write) permissions
		end

	create_with_mode (a_path: STRING; flags, mode: INTEGER) is
			-- create a file according to `flags' and with `mode' access
			-- permissions. Make sure you have th O_CREAT flag in flags
			-- if you really want to create something!
		require
			closed: is_closed
		do
			make_path
			do_create (a_path, flags, mode)
		ensure
			is_opened: security.error_handling.exceptions_enabled implies is_open
		end


feature -- Special creation

	attach_to_fd (a_fd: INTEGER) is
			-- Create file descriptor with value `a_fd'. File descriptor
			-- does not become owner, so it will not close `a_fd'.
		require
			closed: is_closed
			valid_fd: a_fd >= 0
			-- `a_fd' is open
		do
			make_path
			set_fd (a_fd, False)
		ensure
			opened: is_open
			not_owner: not is_owner
			open_files_unchanged:
				security.files.open_files = old security.files.open_files
		end

	make_as_duplicate (another: ABSTRACT_FILE_DESCRIPTOR) is
			-- On creation, create a duplicate from another file descriptor
			-- As normal call, closes its own descriptor first (if open) and
			-- duplicates next.
		local
			a_fd: INTEGER
		do
			make_path
			if is_open then
				a_fd := abstract_dup2 (another.value, fd)
			else
				a_fd := abstract_dup (another.value)
			end
			if a_fd = -1 then
				raise_posix_error
			end
			set_fd (a_fd, True)
		ensure
			open: security.error_handling.exceptions_enabled implies is_open
			own_descriptor: security.error_handling.exceptions_enabled implies is_owner
			open_files_increased_by_one:
				is_owner implies
					security.files.open_files = old security.files.open_files + 1
		end


feature {NONE} -- Open

	do_open (a_path: STRING; flags: INTEGER) is
		require
			closed: is_closed
			-- no_create_flag: flags && O_CREAT = 0
		local
			cpath: POINTER
			a_fd: INTEGER
		do
			set_portable_path (a_path)
			cpath := sh.string_to_pointer (path)
			a_fd := abstract_open (cpath, flags)
			if a_fd = -1 then
				raise_posix_error
			end
			set_fd (a_fd, True)
		ensure
			is_opened: security.error_handling.exceptions_enabled implies is_open
			open_files_increased_by_one:
				is_owner implies
					security.files.open_files = old security.files.open_files + 1
		end

	do_create (a_path: STRING; flags, mode: INTEGER) is
		require
			closed: not is_open
		local
			cpath: POINTER
			a_fd: INTEGER
		do
			set_portable_path (a_path)
			cpath := sh.string_to_pointer (path)
			a_fd := abstract_create (cpath, flags, mode)
			if a_fd = -1 then
				raise_posix_error
			end
			set_fd (a_fd, True)
		ensure
			is_opened: security.error_handling.exceptions_enabled implies is_open
			open_files_increased_by_one:
				is_owner implies
					security.files.open_files = old security.files.open_files + 1
		end


feature -- Close

	close is
			-- We always describe an existing object, however user
			-- probably wants to have control about closing a file.
		require
			opened: is_open
		do
			if is_owner then
				safe_call (abstract_close (fd))
			end
			detach
		ensure
			closed: is_closed
			open_files_decreased_by_one:
				old is_owner implies
					security.files.open_files = old security.files.open_files - 1
		end

	detach is
			-- Forget the current file descriptor.
		do
			clear_fd
			is_owner := False
		ensure
			closed: not is_open
			not_owner: not is_owner
			open_files_decreased_by_one:
				old is_owner implies
					security.files.open_files = old security.files.open_files - 1
		end

	unattach is
		obsolete "Please use detach instead."
		do
			detach
		end


feature -- Change ownership of the descriptor. Can help to influence subtile garbage collector problems

	make_owner is
			-- this file descriptor will (start to) own its descriptor
		do
			is_owner := True
			if is_owner then
				security.files.increase_open_files
			end
		ensure
			owner: is_owner
			open_files_increased_by_one:
				is_owner implies
					security.files.open_files = old security.files.open_files + 1
		end

	unown is
			-- When a stream is opened on a file descriptor the file
			-- descriptor itself should not close itself, the stream
			-- will close it.
		do
			if is_owner then
				security.files.decrease_open_files
			end
			is_owner := False
		ensure
			not_owner: not is_owner
			open_files_decreased_by_one:
				old is_owner implies
					security.files.open_files = old security.files.open_files - 1
		end


feature -- Stream or file behaviour

	is_streaming: BOOLEAN
			-- Is data from this file descriptor coming through a network
			-- stream?

	set_streaming (enable: BOOLEAN) is
			-- Influence behaviour of certain functions if they should be
			-- optimized for data coming from disk or data coming from
			-- the network. In particular `is_streaming' implies that a
			-- client application is prepared to handle `read's that
			-- return less than the requested number of bytes, but don't
			-- assume that means end-of-file.
		do
			is_streaming := enable
			if not (status.is_character_special or else status.is_fifo) then
				read_loop_disabled := is_streaming
			end
		ensure
			streaming: is_streaming = enable
		end


feature -- Raw read and write

	last_blocked: BOOLEAN
			-- True if last read or write call would be blocked

	last_read: INTEGER
			-- How many bytes were read by last call to `read'
			-- -1 implies last_blocked

	last_written: INTEGER
			-- How many bytes were written by last call to write
			-- -1 implies last_blocked

	read_loop_disabled: BOOLEAN
			-- For data coming from the network, a read does not always
			-- return the requested number of bytes. In such a case,
			-- reading again probably will return more bytes as they have
			-- arrived in the network buffers, or perhaps by reading you
			-- have freed the network buffers for more data.
			-- However, certain file descriptors block when reading
			-- again, because they don't return EOF when there is no more
			-- data. A typical example is a character special file.
			-- And sometimes the application can handle streaming reads
			-- just fine and doesn't care if not all requested bytes are
			-- returned.
			-- This variable influences if `read' will attempt to read
			-- more bytes in its loop, or not.

	read (buf: POINTER; offset, nbytes: INTEGER) is
			-- Read data into `buf' at `offset' for `nbytes' bytes.
			-- Number of bytes actually read are available in last_read.
			-- Don't mix this routine with read_string or read_character!
		require
			can_read: is_open
			valid_buf: buf /= default_pointer
			valid_size: nbytes >= 0
			-- `nbytes' < SSIZE_MAX (POSIX value?)
		local
			ptr: POINTER
			nread,
			nleft: INTEGER
		do
			eof := False
			last_read := 0
			last_blocked := False
			errno.clear
			if nbytes > 0 then
				from
					nread := 0
					nleft := nbytes
				invariant
					nleft = 0 or else nbytes = nleft + last_read
				variant
					nleft
				until
					nleft = 0 or else last_blocked
				loop
					ptr := posix_pointer_add (buf, offset + last_read)
					nread := abstract_read (fd, ptr, nleft)
					if nread = -1 then -- test if error occurred
						if errno.value = abstract_EAGAIN then -- test if blocked
							nleft := 0
							-- Set blocked only if nothing read, else return
							-- with data just read.
							last_blocked := last_read = 0
						elseif errno.value = abstract_EINTR then -- test if interrupted
							nread := 0 -- call read() again
						else
							raise_posix_error
							-- In case exceptions are disabled, exit loop.
							nleft := 0
						end
					else
						if nread = 0 then
							-- Set eof only when nothing read the first time.
							eof := nleft = nbytes
							-- Satisfy loop invariant, finish loop.
							nleft := 0
						else
							last_read := last_read + nread
							if read_loop_disabled then
								-- Do not read more than what is available,
								-- else application will block.
								nleft := 0
							else
								nleft := nleft - nread
							end
						end
					end
				end
			end
		ensure
			read_no_more_than_requested: last_read <= nbytes
		end

	write (buf: POINTER; offset, nbytes: INTEGER) is
			-- write given data from `buf' at `offset', for `nbytes' bytes.
		require
			can_write: is_open
			valid_buf: buf /= default_pointer
			valid_size: nbytes >= 0
		local
			ptr: POINTER
			nwritten,
			nleft: INTEGER
		do
			eof := False
			last_written := 0
			last_blocked := False
			errno.clear
			if nbytes > 0 then
				from
					nwritten := 0
					nleft := nbytes
				invariant
					nleft = 0 or else nbytes = nleft + last_written
				variant
					nleft
				until
					nleft = 0 or else last_blocked
				loop
					ptr := posix_pointer_add (buf, offset + last_written)
					nwritten := abstract_write (fd, ptr, nleft)
					if nwritten = -1 then
						if errno.value = abstract_EAGAIN then -- test if blocked
							nleft := 0
							-- set blocked only if nothing read, else return
							-- with data just read
							last_blocked := last_read = 0
						elseif errno.value = abstract_EINTR then -- test if interrupted
							nwritten := 0 -- call write() again
						else
							raise_posix_error
							-- In case exceptions disabled, exit loop.
							nleft := 0
						end
					else
						eof := nwritten = 0
						if eof then
							nleft := 0 -- satisfy loop invariant, finish loop
						else
							nleft := nleft - nwritten
							last_written := last_written + nwritten
						end
					end
				end
			end
		ensure
			wrote_no_more_than_requested: last_written <= nbytes
		end


feature -- Safer read/write

	read_buffer (buf: STDC_BUFFER; offset, nbytes: INTEGER) is
			-- more safe version of `read' in case you have a
			-- STDC_BUFFER object
		require
			dont_overflow: buf.capacity >= offset + nbytes
		do
			read (buf.ptr, offset, nbytes)
		end

	write_buffer (buf: STDC_BUFFER; offset, bytes: INTEGER) is
			-- more safe version of `write' in case you have a
			-- STDC_BUFFER object
		require
			dont_write_garbage: buf.is_valid_index (offset + bytes - 1)
		do
			write (buf.ptr, offset, bytes)
		end


feature -- Read routines that can be mixed with `read' calls

	last_line: STRING
			-- last line read by `read_line (includes %N), see STRING_HELPER.chop

	read_line (max_length: INTEGER) is
			-- Raw, and slow, read of characters up to end of line. Can
			-- be safely mixed with `read', unlike `read_string'. Only
			-- `max_length' characters are returned.
		require
			can_read: is_open
			max_length_not_negative: max_length >= 0
		do
			if last_line = Void then
				last_line := sh.make_with_capacity (max_length)
			end
			from
				last_line.wipe_out
				read ($last_character, 0, 1)
			until
				last_character = '%N' or else
				last_line.count = max_length or else
				eof
			loop
				last_line.append_character (last_character)
				read ($last_character, 0, 1)
			end
			if not eof and last_line.count < max_length then
				last_line.append_character (last_character)
			end
		ensure
			last_line_not_void: last_line /= Void
			last_line_is_limited: last_line.count <= max_length
		end


feature -- Eiffel like output

	put (a: ANY) is
			-- write any Eiffel object as string
		do
			if a /= Void then
				write_string (a.out)
			end
		end

	write_character (c: CHARACTER) is
		do
			put (c)
		end

	write_string, puts, put_string (s: STRING) is
		local
			p: POINTER
		do
			p := sh.string_to_pointer (s)
			write (p, 0, s.count)
		end


feature -- Buffered input, line reading instead of block reading, don't mix with raw `read' calls!

	last_character: CHARACTER

	last_string: STRING
			-- last read string (includes %N), see STRING_HELPER.chop

	read_character is
			-- Sets `last_character'.
		require
			can_read: is_open
		local
			refill_buffer: BOOLEAN
		do
			assert_have_line_buffer
			refill_buffer := line_buffer_index >= line_buffer.count
			if refill_buffer then
				read_buffer (line_buffer, 0, line_buffer.capacity)
				line_buffer.set_count (last_read)
				line_buffer_index := 0
			end
			if line_buffer_index < line_buffer.count then
				last_character := line_buffer.peek_character (line_buffer_index)
				line_buffer_index := line_buffer_index + 1
			else
				last_character := '%U'
			end
		ensure
			last_character_reset_on_eof: eof implies last_character = '%U'
		end

	read_string (a_size: INTEGER) is
			-- Implements line reading on top of `read'. Sets
			-- `last_string' which includes the new line character if
			-- any. Reads until eof or new line encountered, returns max
			-- `a_size' characters.
		require
			can_read: is_open
			valid_size: a_size >= 0
		local
			i: INTEGER
			save_read_loop_disabled: BOOLEAN
		do
			-- make sure last_string has at least `a_size' capacity
			if last_string = Void then
				if last_read > 1024 then
					last_string := sh.make_with_capacity (a_size)
				else
					last_string := sh.make_with_capacity (1024)
				end
			else
				last_string.wipe_out
			end

			assert_end_of_line_read

			-- Read up to newline or a_size or eof, whatever comes first.
			-- Routine probably can be improved somewhat by reading from
			-- line buffer directly and copying data in blocks instead of
			-- character by character.
			-- Because we're reading lines, no need to do a repeating
			-- read in read to get stuff from any kernel buffer, because
			-- we'll come to that later anyway.
			if a_size > 0 then
				save_read_loop_disabled := read_loop_disabled
				read_loop_disabled := True
				from
					i := 0
					last_character := '%U'
				until
					last_character = '%N' or else i = a_size or else
					eof or else last_blocked
				loop
					read_character
					if (not eof) and (not last_blocked) then
						last_string.append_character (last_character)
					end
					i := i + 1
				end
				read_loop_disabled := save_read_loop_disabled
				if last_character /= '%N' then
					line_buffer_skip_rest := True
				end
				if eof and then not last_string.is_empty then
					eof := False
				end
			end
		ensure
			eof_only_when_nothing_read:
				(not last_blocked) implies last_string.is_empty = eof
		end


feature {NONE} -- buffered reading support

	line_buffer: EPX_SMART_BUFFER

	line_buffer_index: INTEGER

	line_buffer_skip_rest: BOOLEAN

	assert_have_line_buffer is
		do
			if line_buffer = Void then
				create line_buffer.allocate (1024)
			end
		end

	assert_end_of_line_read is
			-- Call when read_string has truncated the line and we still
			-- need to read a %N.
		do
			if line_buffer_skip_rest then
				from
					read_character
				until
					eof or else
					last_blocked or else
					last_character = '%N'
				loop
					read_character
				end
				line_buffer_skip_rest := False
			end
		ensure
			skip_reset: not line_buffer_skip_rest
		end


feature -- file position

	seek (offset: INTEGER) is
			-- set file position to given absolute `offset'
		require
			valid_offset: offset >= 0
		do
			safe_call (abstract_lseek (fd, offset, SEEK_SET))
		end

	seek_from_current (offset: INTEGER) is
			-- set file position relative to current position
		do
			safe_call (abstract_lseek (fd, offset, SEEK_CUR))
		end

	seek_from_end (offset: INTEGER) is
			-- set file position relative to end of file
		require
			valid_offset: offset <= 0
		do
			safe_call (abstract_lseek (fd, offset, SEEK_END))
		end


feature -- queries

	eof: BOOLEAN
			-- True if end-of-file reached.
			-- Currently I'm unsure if detection is reliable

	isatty, is_attached_to_terminal: BOOLEAN is
			-- return true if handle associated with character device
		do
			Result := abstract_isatty (fd)
		end

	is_blocking_io: BOOLEAN is
			-- True if blocking i/o enabled (default)
		do
			Result := True
		end

	is_closed: BOOLEAN is
			-- Is file descriptor is closed?
		do
			Result := not is_open
		ensure
			in_balance: Result implies not is_open
		end

	is_open: BOOLEAN
			-- Does `value' describes a valid file descriptor?

	is_owner: BOOLEAN
			-- Does this file descriptor own its descriptor?  Only when
			-- it owns the descriptor it will close it when `close' or
			-- `dispose' is called

	status: ABSTRACT_STATUS is
			-- The status for this file descriptor. Cached value,
			-- refreshed only when file reopened.
		do
			if my_status = Void then
				make_status
			end
			Result := my_status
		end

	value: INTEGER is
			-- return the value of the file descriptor
		require
			valid_file_descriptor: is_open
		do
			Result := fd
		end


feature -- Access

	fd: INTEGER
			-- The actual file descriptor.


feature {NONE} -- Implementation

	my_status: ABSTRACT_STATUS
			-- Cached status object.


	my_buffer: STDC_BUFFER
			-- Buffer used by some read routines.

	assert_valid_buffer (a_capacity: INTEGER) is
			-- Make sure `my_buffer' can contain at least `a_capacity'
			-- elements.
		require
			a_capacity >= 0
		do
			if my_buffer = Void then
				create my_buffer.allocate (a_capacity)
			else
				if my_buffer.capacity < a_capacity then
					create my_buffer.allocate (a_capacity)
				end
			end
		ensure
			buffer_large_enough: my_buffer.capacity >= a_capacity
		end

	clear_fd is
			-- Only allowed routine that may reset a file descriptor.
		do
			if is_open and is_owner then
				security.files.decrease_open_files
			end
			set_fd (-1, False)
		ensure
			closed: not is_open
			not_owner: not is_owner
			fd_reset: fd = -1
			open_files_decreased_by_one:
				old is_owner implies
					security.files.open_files = old security.files.open_files - 1
		end

	dispose is
		local
			r: INTEGER
		do
			if is_open then
				r := abstract_close (fd)
				decrease_open_files
			end
			precursor
		end

	make_status is
			-- give `my_status' a proper value
		deferred
		ensure
			status_is_set: my_status /= Void
		end

	set_fd (a_fd: INTEGER; a_owner: BOOLEAN) is
			-- Only allowed routine that may set a file descriptor value
			-- and perform necessary initializations. Also increments the
			-- open file count if the file descriptor got a value, and
			-- this object is owner.
		require
			valid_fd: a_fd >= -1
		do
			fd := a_fd
			is_open := a_fd >= 0
			is_owner := is_open and a_owner
			if is_owner then
				security.files.increase_open_files
			end
			eof := False
			my_status := Void
			if is_open then
				-- Attempt to determine when to avoid blocking.
				read_loop_disabled :=
					status.is_character_special or else
					status.is_fifo
			end
		ensure
			fd_set: fd = a_fd
			owner_set: is_open implies is_owner = a_owner
			open_files_increased_by_one:
				is_owner implies
					security.files.open_files = old security.files.open_files + 1
			open_files_unchanged:
				not is_owner implies
					security.files.open_files = old security.files.open_files
			eof_reset: not eof
			status_reset: not is_open implies my_status = Void
		end


feature {NONE} -- Abstract API binding

	abstract_close (fildes: INTEGER): INTEGER is
			-- Closes a a file
		deferred
		end

	abstract_create (a_path: POINTER; oflag, mode: INTEGER): INTEGER is
			-- Creates a new file or rewrites an existing one
			-- does not call `creat' but `open'!
		deferred
		end

	abstract_dup (fildes: INTEGER): INTEGER is
			-- Duplicates an open file descriptor
		deferred
		end

	abstract_dup2 (fildes, fildes2: INTEGER): INTEGER is
			-- Duplicates an open file descriptor.
			-- fildes is the file descriptor to duplicate, fildes2 is the
			-- desired new file descriptor.
		deferred
		end

	abstract_isatty (fildes: INTEGER): BOOLEAN is
			-- Determines if a file descriptor is associated with a terminal
		deferred
		end

	abstract_lseek (fildes: INTEGER; offset, whence: INTEGER): INTEGER is
			-- Repositions read/write file offset
		deferred
		end

	abstract_open (a_path: POINTER; oflag: INTEGER): INTEGER is
			-- Opens a file
		deferred
		end

	abstract_read (fildes: INTEGER; buf: POINTER; nbyte: INTEGER): INTEGER is
			-- Reads from a file
		deferred
		end

	abstract_write (fildes: INTEGER; buf: POINTER; nbyte: INTEGER): INTEGER is
			-- Reads from a file
		require
			valid_fildes: fildes >= 0
			valid_buf: buf /= default_pointer
			valid_bytes: nbyte >= 0
		deferred
		end


feature {NONE} -- error codes

	abstract_EAGAIN: INTEGER is
			-- The process would be delayed in the I/O operation
		deferred
		end

	abstract_EINTR: INTEGER is
			-- read/write was interrupted
		deferred
		end


feature {NONE} -- open constants

	abstract_O_APPEND: INTEGER is
			-- Set the file offset to the end-of-file prior to each write
		deferred
		end

	abstract_O_CREAT: INTEGER is
			-- If the file does not exist, allow it to be created. This
			-- flag indicates that the mode argument is present in the
			-- call to open.
		deferred
		end

	abstract_O_RDONLY: INTEGER is
			-- Open for reading only
		deferred
		end

	abstract_O_RDWR: INTEGER is
			-- Open fo reading and writing
		deferred
		end

	abstract_O_TRUNC: INTEGER is
			-- Use only on ordinary files opened for writing. It causes
			-- the file to be truncated to zero length.
		deferred
		end

	abstract_O_WRONLY: INTEGER is
			-- Open for writing only
		deferred
		end


feature {NONE} -- permission constants

	abstract_S_IREAD: INTEGER is
		deferred
		end

	abstract_S_IWRITE: INTEGER is
		deferred
		end


feature {NONE} -- Counting of allocated resource

	decrease_open_files is
			-- Decreate the number of open files, does not access any
			-- other object, so can be called from `dispose' if one
			-- inherits from this class.
		external "C"
		alias "posix_decrease_open_files"
		end


invariant

	valid_internal_file_descriptor: fd >= -1
	valid_open: is_open implies fd >= 0
	valid_close: not is_open implies fd = -1
	valid_status: is_closed implies my_status = Void
	open_and_close_in_balance: is_open = not is_closed
	owner_implies_open: is_owner implies is_open

end
