%{
indexing

	description: "eposix's mime parser."

	standards: "Based on RFC 822, 2045 and 2047."

	implementation:
		"1. Not all recognized field names are parsed.%
		%   Should be easy to expand if you wish to do real message parsing.%
		%2. RFC821 defines maximum line lengths of 1000 characters. This parser%
		%   doesn't care.%
		%3. Fields to add: Content-Type-Encoding; Content-Description"

	author: "Berend de Boer"
	date: "$Date: $"
	revision: "$Revision: $"

class

	EPX_MIME_PARSER


inherit

	EPX_MIME_SCANNER
		rename
			make as make_scanner
		redefine
			wrap
		end

	YY_PARSER_SKELETON [ANY]
		rename
			parse as do_parse
		redefine
			make,
			report_error
		end


creation

	make,
	make_from_file,
	make_from_file_descriptor


%}

-- structured field names
%token FN_CONTENT_DISPOSITION FN_CONTENT_LENGTH FN_CONTENT_TYPE FN_DATE
%token FN_FROM FN_MESSAGE_ID
%token FN_RECEIVED FN_RETURN_PATH FN_RESENT_FROM FN_RESENT_REPLY_TO
%token FN_RESENT_SENDER
%token FN_SENDER FN_TO FN_VERSION

-- structured field elements
%token <STRING> ATOM DOMAIN_LITERAL QUOTED_STRING

-- special keywords to be expected at certain points
%token KW_BY KW_FOR KW_FROM KW_ID KW_MULTIPART KW_VIA KW_WITH
%token KW_SUN KW_MON KW_TUE KW_WED KW_THU KW_FRI KW_SAT

-- encoded words
%token ENCODED_TEXT ENCODED_WORD_START ENCODED_WORD_STOP

-- unstructured fields
%token <STRING> FIELD_NAME FIELD_BODY

-- syntactic elements
%token CRLF NULL_LINE

-- types of nonterminals
%type <EPX_MIME_FIELD> header_field
%type <EPX_MIME_UNSTRUCTURED_FIELD> unrecognized_field
%type <EPX_MIME_FIELD> recognized_field

-- types of recognized fields
%type <EPX_MIME_FIELD_CONTENT_DISPOSITION> content_disposition_field
%type <EPX_MIME_FIELD_CONTENT_TYPE> content_type_field
%type <EPX_MIME_FIELD_CONTENT_LENGTH> content_length_field
%type <EPX_MIME_FIELD_MESSAGE_ID> message_id_field
%type <EPX_MIME_FIELD_MIME_VERSION> mime_version_field

-- types of structured field parts
%type <STRING> addr_spec
%type <STRING> attribute
%type <STRING> domain
%type <STRING> domain_ref
%type <STRING> local_part
%type <STRING> msg_id
%type <EPX_MIME_PARAMETER> parameter
%type <STRING> sub_domain
%type <STRING> subtype
%type <STRING> token
%type <STRING> type
%type <STRING> value

-- basic types handled by nonterminals
%type <INTEGER> integer
%type <STRING> word

%%
-------------------------------------------------------------------------------

message
	: header_fields NULL_LINE { accept }
	| error NULL_LINE { accept }
	;


header_fields
	: header_field
	| header_field header_fields
	;


-- @@BdB error handling incomplete for last field, then we get a NULL_LINE
-- how to handle this?
-- I hope we fall back to the message error recovery, see above
header_field
	: recognized_field
		{ if $1 /= Void then part.fields.put ($1, $1.name) end }
	| unrecognized_field
		{ part.fields.put ($1, $1.name) }
	| error { reset_start_condition } CRLF
		{ stop_keyword }
	;


recognized_field
	: content_disposition_field
		{ $$ := $1 }
	| content_length_field
		{ $$ := $1 }
	| content_type_field
		{ $$ := $1 }
	| date_field
	| from_field
	| message_id_field
		{ $$ := $1 }
	| mime_version_field
		{ $$ := $1 }
	| received_field
	| resent_from_field
	| resent_sender_field
	| resent_reply_to_field
	| return_path_field
	| sender_field
	| to_field
	;


-- structured fields

content_disposition_field
	: FN_CONTENT_DISPOSITION ':'
		{ start_keyword }
	token
		{
			create my_content_disposition.make ($4)
			current_parameter_field := my_content_disposition
		}
	optional_parameters
		{
			$$ := my_content_disposition
			current_parameter_field := Void
		}
	;

content_length_field
	: FN_CONTENT_LENGTH ':'
		{ start_keyword }
	integer
		{
			create $$.make ($4)
		}
	;

content_type_field
	: FN_CONTENT_TYPE ':'
		{ start_keyword }
	type '/' subtype
		{
			create my_content_type.make ($4, $6)
			current_parameter_field := my_content_type
		}
	optional_parameters
		{
			$$ := my_content_type
			current_parameter_field := Void
		}
	;

date_field
	: FN_DATE ':'
		{ start_keyword }
	date_time
	;

from_field
	: FN_FROM ':' mailbox_list
	;

message_id_field
	: FN_MESSAGE_ID ':' msg_id
		{ create $$.make ($3) }
	;

mime_version_field
	: FN_VERSION ':'
		{ start_keyword }
	INTEGER '.' INTEGER
		{ create $$.make ($4, $6) }
	;

received_field
	: FN_RECEIVED ':'
		{ start_keyword }
	optional_from_domain
	optional_by_domain
	optional_via_atom
	optional_with_atoms
	optional_id
	optional_for_addr_spec
	';'{ start_keyword } date_time
	;

resent_from_field
	: FN_RESENT_FROM ':' mailbox_list
	;

resent_reply_to_field
	: FN_RESENT_REPLY_TO ':' address_list
	;

resent_sender_field
	: FN_RESENT_SENDER ':' mailbox
	;

return_path_field
	: FN_RETURN_PATH ':' route_addr
	;

sender_field
	: FN_SENDER ':' mailbox
	;

to_field
	: FN_TO ':' address_list
	;

-- structured fields parts

address
	: mailbox
	| group
	;

address_list
	: ','
	| address
	| address ',' address_list
	;

addr_spec
	: local_part '@' domain
		{
		create $$.make_from_string ($1)
		$$.append_character ('@')
		$$.append_string ($1)
		}
	;

attribute
	: token
		{ $$ := $1 }
	;

charset
	: token
	;

date
	: ATOM ATOM ATOM
	;

date_time
	: optional_day date time
	;

day
	: KW_SUN
	| KW_MON
	| KW_TUE
	| KW_WED
	| KW_THU
	| KW_FRI
	| KW_SAT
	|
	;

domain
	: sub_domain
		{ $$ := $1 }
	| sub_domain '.' domain
		{ $$ := $1 + "." + $3 }
	;

domain_ref
	: atom
		{ $$ := $1 }
	;

encoded_word
	: ENCODED_WORD_START charset '?' encoding '?'
		{ start_encoded_text; }
	ENCODED_TEXT ENCODED_WORD_STOP
	;

encoding
	: token
	;

group
	: phrase ':' optional_mailbox_list ';'
	;

hour
	: ATOM ':' ATOM optional_seconds
	;

integer
	: ATOM
		{
		if $1.is_integer then
			$$ := $1.to_integer
		else
			abort
		end
		}
	;

local_part
	: word
		 { $$ := $1 }
	| word '.' local_part
		 { $$ := $1 + "." + $3 }
	;

mailbox
	: addr_spec
	| phrase route_addr
	;

mailbox_list
	: ','
	| mailbox
	| mailbox ',' mailbox_list
	;

msg_id
	: '<' addr_spec '>'
		{ $$ := $2 }
	;

optional_by_domain
	: -- empty
	| KW_BY domain
	;

optional_day
	: -- empty
	| day ','
		{ stop_keyword }
	;

optional_domain_literal
	: -- empty
	| DOMAIN_LITERAL
	;

-- The angle brackets are not according to the specs
-- but postfix insist on inserting it...
optional_for_addr_spec
	: -- empty
	| KW_FOR addr_spec
	| KW_FOR '<' addr_spec '>'
	;

-- fetchmail inserts a domain literal after from domain
-- ACM Email Forwarding Service has a from with only a comment
optional_from_domain
	: -- empty
	| KW_FROM domain optional_domain_literal
	;

optional_id
	: -- empty
	| KW_ID atom
	;

optional_mailbox_list
	: -- empty
	| mailbox_list
	;

optional_parameters
	: -- empty
	| parameters
	;

optional_route
	: -- empty
	| route
	;

optional_seconds
	: -- empty
	| ':' ATOM
	;

optional_via_atom
	: -- empty
	| KW_VIA atom
	;

optional_with_atoms
	: -- empty
	| with_atoms
	;

path_relative
	: '@' domain ':'
	;

parameter
	: attribute '=' value
		{
			$1.to_lower
			create $$.make ($1, $3)
			current_parameter_field.parameters.put ($$, $1)
		}
	;

parameters
	: ';' parameter
	| ';' parameter parameters
	;

phrase
	: word_or_encoded_word
	| word_or_encoded_word phrase
	;

route
	: ','
	| path_relative
	| path_relative ',' route
	;

route_addr
	: '<' optional_route addr_spec '>'
	;

sub_domain
	: domain_ref
	{ $$  := $1 }
	| DOMAIN_LITERAL
		{ $$ := $1 }
	;

subtype
	: atom
		{ $$ := $1 }
	;

time
	: hour zone
	;

-- content-type to make parsing easier.
token
	: ATOM
		{ $$ := $1 }
	;

-- discrete versus composite type does not seem to have much meaning
-- while parsing.
type
	: KW_MULTIPART
		{
			part.create_multipart_body
			$$ := "multipart"
		}
	| ATOM
		{
			$$ := $1
		}
	;

value
	: token
		{ $$ := $1 }
	| QUOTED_STRING
		{ $$ := $1 }
	;

with_atoms
	: KW_WITH atom
	| KW_WITH atom with_atoms
	;

word
	: ATOM
		{ $$ := $1 }
	| QUOTED_STRING
		{ $$ := $1 }
	;

word_or_encoded_word
	: word
	| encoded_word
	;

-- parse the recognized zones in RFC822??
-- either something like UT or EST
-- or +0400 or -1200
zone
	: ATOM
	;

-- all other fields

unrecognized_field
	: FIELD_NAME ':' FIELD_BODY
		{ create $$.make ($1, $3) }
	;


-------------------------------------------------------------------------------
%%


feature -- Initialization

	make is
		do
			if last_line = Void then
				make_scanner
				precursor
				last_line := STRING_.make (128)
			end
			create part.make_empty
			level := 1
			boundary := Void
		end

	make_from_file (a_file: STDC_TEXT_FILE) is
		require
			file_not_void: a_file /= Void
			file_open: a_file.is_open
		do
			make
			a_file.set_no_buffering
			set_input_buffer (new_stdc_file_buffer (a_file))
		end

	make_from_file_descriptor (a_fd: ABSTRACT_FILE_DESCRIPTOR) is
		require
			file_descriptor_not_void: a_fd /= Void
			file_descriptor_open: a_fd.is_open
		do
			make
			set_input_buffer (new_file_descriptor_buffer (a_fd))
		end


feature -- Character reading

	eof: BOOLEAN is
			-- True if `read_character' hits end-of-file.
		do
			Result :=
				last_character = yyEnd_of_file_character and
				not input_buffer.filled
		end


feature {NONE} -- Scanning

	regular_buffer: YY_BUFFER
			-- Cache of `input_buffer', to be copied back on `wrap'.

	wrap: BOOLEAN is
			-- Check if we were parsing header, if so, resume with main
			-- buffer.
		do
			if regular_buffer /= Void then
				set_input_buffer (regular_buffer)
				regular_buffer := Void
			else
				Result := True
			end
		end


feature -- Access

	token_name (a_token: INTEGER): STRING is
			-- Name of token `a_token'
		do
			Result := mime_tokens.token_name (a_token)
		end


feature -- Parsing

	parse is
		local
			is_multipart_body: BOOLEAN
			multipart_body: EPX_MIME_BODY_MULTIPART
			save_part: EPX_MIME_PART
		do
			do_parse

			-- Be very careful in determining if we should parse a
			-- multipart body.
			is_multipart_body :=
				part.body /= Void and then part.body.is_multipart and then
				part.content_type /= Void and then part.content_type.parameters.has (part.parameter_name_boundary)
			if is_multipart_body then
				part.content_type.parameters.search (part.parameter_name_boundary)
				is_multipart_body := part.content_type.parameters.found
				if is_multipart_body then
					boundary := "--" + part.content_type.parameters.found_item.value
					is_multipart_body := boundary.count <= Max_rfc_2046_boundary_length + 2
				end
			end

			if is_multipart_body then
				-- some overflow test here in `level'?
				save_part := part
				multipart_body ?= save_part.body
				level := level + 1
				forward_to_boundary
				from
				until
					end_of_file or else
					eof or else
					boundary_with_trailer_read
				loop
					part := multipart_body.new_part
					parse
				end
				level := level - 1
				part := save_part
			else
				-- We get here to read a body which ends with eof or
				-- a mime boundary. Make sure we have a single part body.
				part.clear_body
				part.create_singlepart_body
				read_singlepart_body
			end
		end

	set_header (a_header: STRING) is
			-- Optional header that is parsed before the regular input
			-- is parsed.
		require
			a_header_not_empty: a_header /= Void and then not a_header.is_empty
		local
			string_buffer: YY_BUFFER
		do
			regular_buffer := input_buffer
			create string_buffer.make (a_header)
			set_input_buffer (string_buffer)
		end


feature {NONE} -- Error reporting

	report_error (a_message: STRING) is
			-- Dump error to stderr. Needs to be rewritten.
		do
			std.error.put_string ("line ")
			std.error.put_integer (line)
			std.error.put_string (", ")
			std.error.put_integer (column)
			std.error.put_string (": ")
			std.error.put_string (a_message)
			std.error.put_character ('%N')
			std.error.put_string ("  text: %"")
			if last_value /= Void then
				std.error.put_string (last_value)
			end
			std.error.put_string ("%" (")
			std.error.put_string (text)
			std.error.put_string (")%N")
			std.error.put_string ("  token: ")
			std.error.put_integer (last_token)
			std.error.put_character ('%N')
		end


feature -- State

	file: STDC_TEXT_FILE
			-- Where data comes from.

	part: EPX_MIME_PART
			-- Structure we're building.


feature {NONE} -- State used during parsing

	level: INTEGER

	my_content_disposition: EPX_MIME_FIELD_CONTENT_DISPOSITION

	my_content_type: EPX_MIME_FIELD_CONTENT_TYPE

	current_parameter_field: EPX_MIME_FIELD_WITH_PARAMETERS


feature {NONE} -- Reading mime bodies

	boundary: STRING
			-- For multipart MIME messages, it designates the end of a body.

	boundary_with_trailer_read: BOOLEAN

	determine_boundary_with_trailer_read is
			-- There was a boundary match, determine if two dashes follow
			-- (last boundary) and set `boundary_with_trailer_read' in that case.
		do
			-- NOTE: we've already read one character.
			boundary_with_trailer_read := last_character = '-'
			if boundary_with_trailer_read then
				read_character
				boundary_with_trailer_read := last_character = '-'
			end
		end

	forward_to_boundary is
			-- Move input cursor to line that contains `boundary'.
			-- Assume we start reading at beginning of a line.
		require
			boundary_not_empty: boundary /= Void and then not boundary.is_empty
		local
			boundary_read: BOOLEAN
			matching_boundary: BOOLEAN
			matched_index: INTEGER
			match: BOOLEAN
		do
			from
				matching_boundary := True
				matched_index := 1
				read_character
			until
				eof or else
				boundary_read
			loop
				if last_character = '%N' then
					-- Start matching beginning of boundary
					matching_boundary := True
					matched_index := 1
				else
					if matching_boundary then
						match := boundary.item (matched_index) = last_character
						if match then
							-- Match found, advance to next character.
							matched_index := matched_index + 1
							boundary_read := matched_index > boundary.count
						else
							-- Mismatch. As boundary has to match only at
							-- beginning of line, we just have a data line.
							matching_boundary := False
						end
					end
				end
				read_character
			end

			if boundary_read then
				determine_boundary_with_trailer_read
			end
			forward_to_end_of_line
		end

	forward_to_end_of_line is
			-- Always move cursor to next line.
		do
			from
			until
				eof or else
				last_character = '%N'
			loop
				read_character
			end
		ensure
			cursor_at_new_line: eof or else last_character = '%N'
		end

	is_text_body: BOOLEAN is
			-- Contains part.body text?
		local
			text_body: EPX_MIME_BODY_TEXT
		do
			text_body ?= part.body
			Result := text_body /= Void
		end

	last_line: STRING
			-- We store last read line here when reading multipart bodies

	Max_rfc_2046_boundary_length: INTEGER is 70
			-- Max length of a boundary according to RFC 2046

	read_singlepart_body is
			-- Read from input until end of file is reached or a line
			-- contains `boundary'.
		require
			body_not_void: part.body /= Void
			body_is_single_part: not part.body.is_multipart
		do
			if boundary = Void then
				read_singlepart_body_without_boundary
			else
				read_singlepart_body_with_boundary
			end
		end

	read_singlepart_body_with_boundary is
			-- Start reading a single part. This routine has to do two
			-- things right: it has to stop when a boundary has been
			-- parsed, and the cr+lf before the boundary is not part of
			-- the file.
		require
			boundary_is_set: boundary /= Void
			has_buffer: last_line /= Void
			body_contains_text: is_text_body
		local
			boundary_read: BOOLEAN
			matched_index: INTEGER
			match: BOOLEAN
			c: CHARACTER
			matching_boundary: BOOLEAN
			body: EPX_MIME_BODY_TEXT
			add_cr: BOOLEAN
		do
			body ?= part.body
			-- This loop reads data character by character.
			-- We have to stop when a boundary occurs after a CRLF.
			from
				matching_boundary := True
				last_line.wipe_out
				matched_index := 1
				read_character
			invariant
				matched_index >= 1
				not boundary_read implies matched_index <= boundary.count
				matching_boundary implies matched_index <= last_line.count + 1
			until
				eof or else
				boundary_read
			loop

				-- `c' contains the current character
				c := last_character

				inspect c
				when '%R' then
					-- when a CR is encountered, we've to check for a LF to
					-- know if this is an end of line or an incidental CR
					-- (pure data)
					last_line.append_character (c)
				when '%N' then
					-- the next line could contain the boundary so we don't
					-- write the (CR)LF, but keep it in `last_line'. The
					-- CRLF is part of the boundary in that case, not of
					-- the message.
					if last_line.count > 0 and then
						last_line.item (last_line.count) = '%R' then
						-- at this point a CRLF has been read
						last_line.remove (last_line.count)
						add_cr := True
					else
						add_cr := False
					end
					body.append_string (last_line)
					last_line.wipe_out
					if add_cr then
						last_line.append_character ('%R')
					end
					last_line.append_character (c)
					matched_index := 1
					matching_boundary := True
				else
					if matching_boundary then
						match := boundary.item (matched_index) = c
						if match then
							-- Match found, advance to next character.
							last_line.append_character (c)
							matched_index := matched_index + 1
							boundary_read := matched_index > boundary.count
						else
							-- Mismatch. As boundary has to match only at
							-- beginning of line, we just have a data line.
							body.append_string (last_line)
							body.append_character (c)
							last_line.wipe_out
							matching_boundary := False
						end
					else
						body.append_string (last_line)
						body.append_character (c)
						last_line.wipe_out
					end
				end

				-- very expensive check
					check
						-- last_line.has_substring (boundary) implies boundary_read
					end

				read_character
			end
			body.finish

			-- if boundary read, skip rest of line
			if boundary_read then
				determine_boundary_with_trailer_read
			end
			forward_to_end_of_line

				check
					stop_condition: boundary_read or eof
				end
		end

	read_singlepart_body_without_boundary is
			-- Read `file' until `end_of_file'. Does handle any line length.
		require
			body_contains_text: is_text_body
		local
			body: EPX_MIME_BODY_TEXT
		do
			body ?= part.body
			from
				read_character
			until
				eof
			loop
				body.append_character (last_character)
				read_character
			end
			body.finish
		ensure
			read_everything: eof
		end


invariant

	last_line_not_void: last_line /= Void

end
